from prov.model import ProvDocument
from provdbconnector import ProvDb
from provdbconnector.db_adapters.in_memory import SimpleInMemoryAdapter

prov_api = ProvDb(adapter=SimpleInMemoryAdapter, auth_info=None)

# create the prov first document
first_prov_document = ProvDocument()
first_prov_document .add_namespace("ex", "http://example.com")

first_prov_document .agent("ex:Bob")
first_prov_document .activity("ex:Alice")

first_prov_document .association("ex:Alice", "ex:Bob")

first_document_id = prov_api.save_document(first_prov_document)

#Create the second prov document and merge the ex:Bob entry
second_prov_document = ProvDocument()
second_prov_document.add_namespace("ex", "http://example.com")

second_prov_document.agent("ex:Bob", other_attributes={"ex:age": 42})

second_document_id = prov_api.save_document(second_prov_document)


#Query the first document ID but get the Bob with the age property back (so successfully merged)

print(prov_api.get_document_as_provn(first_document_id))

# Output:
#
# document
# prefix
# ex < http: // example.com >
#
# activity(ex:Alice, -, -)
# agent(ex:Bob, [ex:age = 42])
# wasAssociatedWith(ex:Alice, ex:Bob, -)
# endDocument



from provdbconnector import ProvDb
from provdbconnector import Neo4jAdapter
import  os
import pkg_resources

# Data is from: https://provenance.ecs.soton.ac.uk/store/documents/75490/


NEO4J_USER = os.environ.get('NEO4J_USERNAME', 'neo4j')
NEO4J_PASS = os.environ.get('NEO4J_PASSWORD', 'neo4jneo4j')
NEO4J_HOST = os.environ.get('NEO4J_HOST', 'localhost')
NEO4J_BOLT_PORT = os.environ.get('NEO4J_BOLT_PORT', '7687')

# Auth info
auth_info = {"user_name": NEO4J_USER,
             "user_password": NEO4J_PASS,
             "host": NEO4J_HOST + ":" + NEO4J_BOLT_PORT
             }


# create the api
prov_api = ProvDb(adapter=Neo4jAdapter, auth_info=auth_info)

# create the prov document from examples
prov_document_buffer = pkg_resources.resource_stream("examples", "horsemeat_example.json")

# Save document
document_id = prov_api.save_document(prov_document_buffer)
# This is similar to:
# prov_api.create_document_from_json(prov_document_buffer)

# get document
print(prov_api.get_document_as_provn(document_id))


from provdbconnector.prov_db import ProvDb

from provdbconnector.exceptions.provapi import ProvDbException
from provdbconnector import db_adapters
from provdbconnector.db_adapters.neo4j.neo4jadapter import Neo4jAdapter
from provdbconnector.db_adapters.neo4j.neo4jadapter import NEO4J_USER, NEO4J_PASS, NEO4J_HOST, NEO4J_HTTP_PORT, NEO4J_BOLT_PORT

from provdbconnector.db_adapters.in_memory import SimpleInMemoryAdapter



from provdbconnector import ProvDb
from provdbconnector.db_adapters.in_memory import SimpleInMemoryAdapter
import pkg_resources

# create the api
prov_api = ProvDb(adapter=SimpleInMemoryAdapter, auth_info=None)

# create the prov document from examples
prov_document_buffer = pkg_resources.resource_stream("examples", "file_buffer_example_primer.json")

# Save document
document_id = prov_api.save_document(prov_document_buffer)
# This is similar to:
# prov_api.create_document_from_json(prov_document_buffer)

# get document
print(prov_api.get_document_as_provn(document_id))

# Output:

# document
# prefix
# foaf < http: // xmlns.com / foaf / 0.1 / >
# prefix
# dcterms < http: // purl.org / dc / terms / >
# prefix
# ex < http: // example / >
#
# specializationOf(ex:articleV2, ex:article)
# specializationOf(ex:articleV1, ex:article)
# wasDerivedFrom(ex:blogEntry, ex:article, -, -, -, [prov:type = 'prov:Quotation'])
# alternateOf(ex:articleV2, ex:articleV1)
# wasDerivedFrom(ex:articleV1, ex:dataSet1, -, -, -)
# wasDerivedFrom(ex:articleV2, ex:dataSet2, -, -, -)
# wasDerivedFrom(ex:dataSet2, ex:dataSet1, -, -, -, [prov:type = 'prov:Revision'])
# used(ex:correct, ex:dataSet1, -)
# used(ex:compose, ex:dataSet1, -, [prov:role = "ex:dataToCompose"])
# wasDerivedFrom(ex:chart2, ex:dataSet2, -, -, -)
# wasGeneratedBy(ex:dataSet2, ex:correct, -)
# used(ex:compose, ex:regionList, -, [prov:role = "ex:regionsToAggregateBy"])
# used(ex:illustrate, ex:composition, -)
# wasGeneratedBy(ex:composition, ex:compose, -)
# wasAttributedTo(ex:chart1, ex:derek)
# wasGeneratedBy(ex:chart1, ex:compile, 2012 - 03 - 02
# T10:30:00)
# wasGeneratedBy(ex:chart1, ex:illustrate, -)
# wasAssociatedWith(ex:compose, ex:derek, -)
# wasAssociatedWith(ex:illustrate, ex:derek, -)
# actedOnBehalfOf(ex:derek, ex:chartgen, ex:compose)
# entity(ex:article, [dcterms:title = "Crime rises in cities"])
# entity(ex:articleV1)
# entity(ex:articleV2)
# entity(ex:dataSet1)
# entity(ex:dataSet2)
# entity(ex:regionList)
# entity(ex:composition)
# entity(ex:chart1)
# entity(ex:chart2)
# entity(ex:blogEntry)
# activity(ex:compile, -, -)
# activity(ex:compile2, -, -)
# activity(ex:compose, -, -)
# activity(ex:correct, 2012 - 03 - 31
# T09:21:00, 2012 - 04 - 01
# T15:21:00)
# activity(ex:illustrate, -, -)
# agent(ex:derek, [foaf:mbox = "<mailto:derek@example.org>", foaf:givenName = "Derek", prov:type = 'prov:Person'])
# agent(ex:chartgen, [foaf:name = "Chart Generators Inc", prov:type = 'prov:Organization'])
# endDocument



from prov.model import ProvDocument
from provdbconnector import ProvDb
from provdbconnector.exceptions.database import MergeException
from provdbconnector.db_adapters.in_memory import SimpleInMemoryAdapter

prov_api = ProvDb(adapter=SimpleInMemoryAdapter, auth_info=None)

# create the prov first document
first_prov_document = ProvDocument()
first_prov_document .add_namespace("ex", "http://example.com")

first_prov_document .agent("ex:Bob", other_attributes={"ex:last_name": "Meier"})
first_prov_document .activity("ex:Alice")

first_prov_document .association("ex:Alice", "ex:Bob")

first_document_id = prov_api.save_document(first_prov_document)

#Create the second prov document and merge the ex:Bob entry
second_prov_document = ProvDocument()
second_prov_document.add_namespace("ex", "http://example.com")

second_prov_document.agent("ex:Bob", other_attributes={"ex:age": 42, "ex:last_name": "Müller"})

try:
    second_document_id = prov_api.save_document(second_prov_document)
except MergeException as e:
    print ("Got the merge exception: {}".format(e))

# Query the first document ID and get only the last_name = Meier back. The merge was not successful.

print(prov_api.get_document_as_provn(first_document_id))

# Output:
#
# document
# prefix
# ex < http: // example.com >
#
# activity(ex:Alice, -, -)
# agent(ex:Bob, [ex:last_name = "Meier"])
# wasAssociatedWith(ex:Alice, ex:Bob, -)
# endDocument


from .provapi import ProvDbException


class ConverterException(ProvDbException):
    """
    Base exception class for document converter.
    """
    pass


class ParseException(ConverterException):
    """
    Thrown, if a given statement could not ne parsed.
    """
    pass


class NoDocumentException(ConverterException):
    """
    Thrown, if no document argument is passed.
    """
    pass


class SerializerException(ProvDbException):
    """
    Base exception class for serializer.
    """
    pass


class ValidatorException(ProvDbException):
    """
    Base exception class for validator.
    """
    pass



import unittest

from provdbconnector.exceptions.database import InvalidOptionsException, AuthException
from provdbconnector import Neo4jAdapter, NEO4J_USER, NEO4J_PASS, NEO4J_HOST, NEO4J_BOLT_PORT
from provdbconnector.prov_db import ProvDb
from provdbconnector.tests import AdapterTestTemplate
from provdbconnector.tests import ProvDbTestTemplate


class Neo4jAdapterTests(AdapterTestTemplate):
    """
    This test extends from AdapterTestTemplate and provide a common set for the neo4j adapter
    """
    def setUp(self):
        """
        Setup the test
        """
        self.instance = Neo4jAdapter()
        auth_info = {"user_name": NEO4J_USER,
                     "user_password": NEO4J_PASS,
                     "host": NEO4J_HOST + ":" + NEO4J_BOLT_PORT
                     }
        self.instance.connect(auth_info)
        session = self.instance._create_session()
        session.run("MATCH (x) DETACH DELETE x")

    @unittest.skip(
        "Skipped because the server configuration currently is set to 'no password', so the authentication will never fail")
    def test_connect_fails(self):
        """
        Try to connect with the wrong password
        """
        auth_info = {"user_name": NEO4J_USER,
                     "user_password": 'xxxxxx',
                     "host": NEO4J_HOST + ":" + NEO4J_BOLT_PORT
                     }
        self.instance.connect(auth_info)
        with self.assertRaises(AuthException):
            self.instance.connect(auth_info)

    def test_connect_invalid_options(self):
        """
        Try to connect with some invalid arguments
        """
        auth_info = {"u": NEO4J_USER,
                     "p": 'xxxxxx',
                     "h": NEO4J_HOST + ":" + NEO4J_BOLT_PORT
                     }
        with self.assertRaises(InvalidOptionsException):
            self.instance.connect(auth_info)

    def tearDown(self):
        """
        Delete all data on the database
        :return:
        """
        session = self.instance._create_session()
        session.run("MATCH (x) DETACH DELETE x")
        del self.instance


class Neo4jAdapterProvDbTests(ProvDbTestTemplate):
    """
    High level api test for the neo4j adapter
    """
    def setUp(self):
        self.auth_info = {"user_name": NEO4J_USER,
                          "user_password": NEO4J_PASS,
                          "host": NEO4J_HOST + ":" + NEO4J_BOLT_PORT
                          }
        self.provapi = ProvDb(api_id=1, adapter=Neo4jAdapter, auth_info=self.auth_info)

    def clear_database(self):
        """
        This function get called before each test starts

        """
        session = self.provapi._adapter._create_session()
        session.run("MATCH (x) DETACH DELETE x")

    def tearDown(self):
        """
        Delete all data in the database
        """
        session = self.provapi._adapter._create_session()
        session.run("MATCH (x) DETACH DELETE x")
        del self.provapi



import json
import logging
import os
from collections import namedtuple
from io import StringIO
from uuid import uuid4

from prov.constants import PROV_ATTRIBUTES, PROV_MENTION, PROV_BUNDLE, PROV_LABEL, PROV_TYPE
from prov.model import ProvDocument, ProvEntity, ProvBundle, ProvRecord, ProvElement, ProvRelation, QualifiedName, \
    ProvAssociation, PROV_REC_CLS, ProvActivity, ProvAgent, PROV_AGENT,PROV_ENTITY,PROV_ACTIVITY, PROV_ATTR_AGENT,PROV_ATTR_ACTIVITY, PROV_ATTR_ENTITY,PROV_ATTR_BUNDLE
from provdbconnector.db_adapters.baseadapter import METADATA_KEY_PROV_TYPE, METADATA_KEY_IDENTIFIER, \
    METADATA_KEY_NAMESPACES, \
    METADATA_KEY_TYPE_MAP, METADATA_KEY_IDENTIFIER_ORIGINAL
from provdbconnector.exceptions.provapi import NoDataBaseAdapterException, InvalidArgumentTypeException, \
    InvalidProvRecordException
from provdbconnector.exceptions.utils import ParseException
from provdbconnector.exceptions.database import NotFoundException
from provdbconnector.utils.converter import form_string, to_json, to_provn, to_xml
from provdbconnector.utils.serializer import encode_json_representation, add_namespaces_to_bundle, create_prov_record, \
    PROV_ATTR_BASE_CLS, serialize_namespace

LOG_LEVEL = os.environ.get('LOG_LEVEL', '')
NUMERIC_LEVEL = getattr(logging, LOG_LEVEL.upper(), None)
logging.basicConfig(level=NUMERIC_LEVEL)
logging.getLogger("prov.model").setLevel(logging.WARN)
log = logging.getLogger(__name__)

PROV_API_BUNDLE_IDENTIFIER_PREFIX = "prov:bundle:{}"


class ProvDb(object):
    """
    The public api class. This class provide methods to save and get documents or part of ProvDocuments

    """

    def __init__(self, api_id=None, adapter=None, auth_info=None, *args):
        """
        Save a new instance of ProvAPI

        :param api_id: The id of the api, optional
        :type api_id: int or str
        :param adapter: The adapter class, must enhance from BaseAdapter
        :type adapter: Baseadapter
        :param auth_info: A dict object that contains the information for authentication
        :type auth_info: dict or None
        """
        if api_id is None:
            self.api_id = uuid4()
        else:
            self.api_id = api_id

        if adapter is None:
            raise NoDataBaseAdapterException()
        self._adapter = adapter()
        self._adapter.connect(auth_info)

    # Converter Methods
    def save_document_from_json(self, content=None):
        """
        Saves a new document in the database

        :param content: The content
        :type content: str or buffer
        :return: document_id
        :rtype: str or buffer
        """
        prov_document = form_string(content=content)
        return self.save_document(content=prov_document)

    def get_document_as_json(self, document_id=None):
        """
        Get a ProvDocument from the database based on the document_id

        :param document_id: document id
        :type document_id: str
        :return: ProvDocument as json string
        :rtype: str
        """
        prov_document = self.get_document_as_prov(document_id=document_id)
        return to_json(prov_document)

    def save_document_from_xml(self, content=None):
        """
        Saves a prov document in the database based on the xml file

        :param content: The content
        :type content: str or buffer
        :return: document_id
        :rtype: str
        """
        prov_document = form_string(content=content)
        return self.save_document(content=prov_document)

    def get_document_as_xml(self, document_id=None):
        """
        Get a ProvDocument from the database based on the document_id

        :param document_id: The id
        :type document_id: str
        :return: ProvDocument as XML string
        :rtype: str
        """
        prov_document = self.get_document_as_prov(document_id=document_id)
        return to_xml(prov_document)

    def save_document_from_provn(self, content=None):
        """
        Saves a prov document in the database based on the provn string or buffer

        :param content: provn object
        :type content: str or buffer
        :return: Document_id
        :rtype: str
        """
        prov_document = form_string(content=content)
        return self.save_document(content=prov_document)

    def get_document_as_provn(self, document_id=None):
        """
        Get a ProvDocument from the database based on the document_id

        :param document_id: The id
        :type document_id: str
        :return: ProvDocument
        :rtype: ProvDocument
        """
        prov_document = self.get_document_as_prov(document_id=document_id)
        return to_provn(prov_document)

    def save_document_from_prov(self, content=None):
        """
        Saves a prov document in the database based on the prov document

        :param content: Prov document
        :type content: ProvDocument
        :return: document_id
        :rtype: str
        """
        if not isinstance(content, ProvDocument):
            raise InvalidArgumentTypeException()
        return self.save_document(content=content)

    # Methods that consume ProvDocument instances and produce ProvDocument instances
    def save_document(self, content=None):
        """
        The main method to Save a document in the db

        :param content: The content can be a xml, json or provn string or buffer or a ProvDocument instance
        :type content: str or buffer or ProvDocument
        :return: Document id
        :rtype: str
        """

        # Try to convert the content into the provDocument, if it is already a ProvDocument instance the function will return this document
        try:
            content = form_string(content=content)
        except ParseException as e:
            raise InvalidArgumentTypeException(e)

        prov_document = content

        doc_id = self._save_bundle_internal(prov_document)

        for bundle in prov_document.bundles:
            self.save_bundle(prov_bundle=bundle)

        return doc_id

    def get_document_as_prov(self, document_id=None):
        """
        Get a ProvDocument from the database based on the document id

        :param document_id: The id
        :type document_id: str
        :return: Prov Document
        :rtype: ProvDocument
        """
        if type(document_id) is not str:
            raise InvalidArgumentTypeException()

        filter_meta = dict()
        filter_prop = dict()
        filter_meta.update({document_id: True})
        filter_prop.update({PROV_TYPE: PROV_BUNDLE})

        bundle_entities = self._adapter.get_records_by_filter(metadata_dict=filter_meta, attributes_dict=filter_prop)
        document_records = self._adapter.get_records_by_filter(metadata_dict=filter_meta)

        # parse document
        prov_document = ProvDocument()
        for record in document_records:
            self._parse_record(prov_document, record)

        bundle_doc = ProvDocument()# Document with all bundle entities

        for bundle_record in bundle_entities:

            # skip if we got some relations instead of only the bundle nodes
            if str(PROV_TYPE) not in bundle_record.attributes:
                continue

            if str(bundle_record.attributes[str(PROV_TYPE)]) != str(PROV_BUNDLE):
                continue

            bundle_entity= self._parse_record(bundle_doc,bundle_record)
            prov_bundle = self.get_bundle(bundle_entity.identifier)
            prov_document.add_bundle(prov_bundle,identifier=bundle_entity.identifier)


        return prov_document

    def save_element(self, prov_element, bundle_id=None):
        """
        Saves a activity, entity, agent

        .. code:: python


            doc = ProvDocument()

            agent       = doc.agent("ex:yourAgent")
            activity    = doc.activity("ex:yourActivity")
            entity      = doc.entity("ex:yourEntity")

            # Save the elements
            agent_id = prov_db.save_element(agent)
            activity_id = prov_db.save_element(activity)
            entity_id = prov_db.save_element(entity)


        :param prov_element: The ProvElement
        :type prov_element: prov.model.ProvElement
        :param bundle_id:
        :type bundle_id: str
        :return: Identifier of the element
        :rtype: prov.model.QualifiedName
        """
        if not isinstance(prov_element, ProvElement):
            raise InvalidArgumentTypeException("Should be {} but was {}".format(ProvElement, type(prov_element)))

        (metadata, attributes) = self._get_metadata_and_attributes_for_record(prov_element, bundle_id=bundle_id)
        self._adapter.save_element(attributes=attributes, metadata=metadata)

        #Add bundle relation only if the record belongs to a bundle not to document
        if not isinstance(prov_element.bundle, ProvDocument):
            bundle_id_qualified = prov_element.bundle.valid_qualified_name(prov_element.bundle.identifier)
            self._create_bundle_association([prov_element], bundle_id_qualified)

        return prov_element.identifier

    def get_elements(self, prov_element_cls):
        """
        Return a document that contains the requested type

        .. code:: python

            from prov.model import ProvEntity, ProvAgent, ProvActivity

            document_with_all_entities = prov_db.get_elements(ProvEntity)
            document_with_all_agents = prov_db.get_elements(ProvAgent)
            document_with_all_activities = prov_db.get_elements(ProvActivity)

            print(document_with_all_entities)
            print(document_with_all_agents)
            print(document_with_all_activities)


        :param prov_element_cls:
        :return: Prov document
        :rtype prov.model.ProvDocument

        """
        if prov_element_cls is ProvAgent:
            prov_type = PROV_AGENT
        elif prov_element_cls is ProvActivity:
            prov_type = PROV_ACTIVITY
        elif prov_element_cls is ProvEntity:
            prov_type = PROV_ENTITY
        else:
            raise InvalidArgumentTypeException("You provide a wrong type : {}".format(type(prov_element_cls)))

        meta_filter = dict()
        meta_filter.update({METADATA_KEY_PROV_TYPE: prov_type})

        doc = ProvDocument()
        raw_results = self._adapter.get_records_by_filter(metadata_dict=meta_filter)

        for element in raw_results:
            if element.metadata[METADATA_KEY_PROV_TYPE] == str(prov_type):
                self._parse_record(doc,element)
        return doc

    def get_element(self, identifier):
        """
        Get a element (activity, agent, entity) from the database

        .. code:: python

            doc = ProvDocument()

            identifier = QualifiedName(doc, "ex:yourAgent")

            prov_element = prov_db.get_element(identifier)


        :param identifier:
        :type identifier: prov.model.QualifiedName
        :return: A prov Element class
        """

        if not isinstance(identifier, QualifiedName):
            raise InvalidArgumentTypeException("Should be {} but was {}".format(QualifiedName, type(identifier)))

        # Include namespace uri into the identifier to support e.g. different default namespaces
        global_identifier = identifier.namespace.uri + identifier.localpart

        # Setup filter
        meta_filter = dict()
        meta_filter.update({METADATA_KEY_IDENTIFIER: global_identifier})

        # Get the result
        results = self._adapter.get_records_by_filter(metadata_dict=meta_filter)

        # Check if there is some unexpected result
        if len(results) > 1:
            raise InvalidProvRecordException("Invalid data result, len should be only one, result was: {}".format(list(results)))
        if len(results) == 0:
            raise NotFoundException("Can't find the element with identifier {}".format(identifier))

        # get the lonely element in the list
        element = list(results).pop()

        doc = ProvDocument()
        return self._parse_record(doc,element)

    def save_record(self, prov_record, bundle_id=None):
        """
        Saves a realtion or a element (Entity, Agent or Activity)

        .. code:: python


            doc = ProvDocument()

            agent       = doc.agent("ex:Alice")
            ass_rel     = doc.association("ex:Alice", "ex:Bob")

            # Save the elements
            agent_id = prov_db.save_record(agent)
            relation_id = prov_db.save_record(ass_rel)

        :param prov_record: The prov record
        :type prov.model.ProvRecord
        :param bundle_id: The bundle id that you got back if you created a bundle or document
        :type str
        :return:
        """
        if not isinstance(prov_record,ProvRecord):
            raise InvalidArgumentTypeException("Wrong type, expected: {}, got {}".format(type(ProvRecord), type(prov_record)))

        if isinstance(prov_record,ProvRelation):
            return self.save_relation(prov_relation=prov_record,bundle_id=bundle_id)
        elif isinstance(prov_record,ProvElement):
            return self.save_element(prov_element=prov_record,bundle_id=bundle_id)
        else:
            raise InvalidArgumentTypeException("Oh no... you provided a not supported prov_record type. The type was: {}".format(type(prov_record)))

    @staticmethod
    def _parse_record(prov_bundle, raw_record):
        """
        This method Saves a ProvRecord in the ProvBundle based on the raw database response

        :param prov_bundle: ProvBundle instance
        :type prov_bundle: ProvBundle
        :param raw_record: DbRelation or DbRecord instance as (namedtuple)
        :type raw_record: DbRelation or DbRecord
        :return ProvRecord
        :rtype prov.model.ProvRecord
        """

        # check if record belongs to this bundle
        prov_type = raw_record.metadata[METADATA_KEY_PROV_TYPE]
        prov_type = prov_bundle.valid_qualified_name(prov_type)

        # skip connections between bundle entities and all records that belong to the bundle
        prov_label = raw_record.attributes.get(str(PROV_LABEL))
        if prov_label is not None and prov_label == "belongsToBundle":
            return

        prov_id = raw_record.metadata[METADATA_KEY_IDENTIFIER_ORIGINAL]
        prov_id_qualified = prov_bundle.valid_qualified_name(prov_id)

        # skip record if prov:identifier starts with "prov:Unknown"
        if str(prov_id).startswith("prov:Unknown-"):
            return

        # set identifier only if it is not a prov type
        if prov_id_qualified == prov_type:
            prov_id = None

        # get type map
        type_map = raw_record.metadata[METADATA_KEY_TYPE_MAP]

        if type(type_map) is str:
            io = StringIO(type_map)
            type_map = json.load(io)

        elif type(type_map) is list:
            type_map_list = type_map
            type_map = dict()
            for type_str in type_map_list:
                if type(type_str) is not str:
                    raise InvalidArgumentTypeException("The type_map must be a string got: {}".format(type_str))
                io = StringIO(type_str)
                type_map.update(json.load(io))

        elif type(type_map) is not dict:
            raise InvalidArgumentTypeException("The type_map must be a dict or json string got: {}".format(type_map))

        add_namespaces_to_bundle(prov_bundle, raw_record.metadata)
        return create_prov_record(prov_bundle, prov_type, prov_id, raw_record.attributes, type_map)

    def get_bundle(self, identifier):
        """
        Returns the whole bundle for the provided identifier

        .. code:: python
            doc = ProvDocument()
            bundle_name = doc.valid_qualified_name("ex:YourBundleName")
            # get the bundle
            prov_bundle = prov_db.get_bundle(bundle_name)
            doc.add_bundle(prov_bundle)


        :param identifier: The identifier
        :type identifier: prov.model.QualifiedName
        :return: The prov bundle instance
        :rtype prov.model.ProvBundle
        """
        if not isinstance(identifier, QualifiedName):
            raise InvalidArgumentTypeException()


        bundle_entity = self.get_element(identifier)

        doc = ProvDocument()
        doc.add_record(bundle_entity)#Add bundle entity to document

        prov_bundle = doc.bundle(identifier=bundle_entity.identifier)

        # Include namespace uri into the identifier to support e.g. different default namespaces
        global_identifier = identifier.namespace.uri + identifier.localpart
        bundle_records = self._adapter.get_bundle_records(global_identifier)

        for record in bundle_records:
            self._parse_record(prov_bundle, record)

        return prov_bundle

    def save_bundle(self,prov_bundle):
        """
        Public method to save a bundle

        .. code:: python

            doc = ProvDocument()

            bundle = doc.bundle("ex:bundle1")
            # Save the bundle
            prov_db.save_bundle(bundle)

        :param prov_bundle:
        :type prov_bundle: prov.model.ProvBundle
        :return:
        """

        if not isinstance(prov_bundle, ProvBundle):
            raise InvalidArgumentTypeException()
        if isinstance(prov_bundle, ProvDocument):
            raise  InvalidArgumentTypeException()

        # create bundle entity
        bundle_record = ProvEntity(prov_bundle.document, identifier=prov_bundle.identifier, attributes={PROV_TYPE: PROV_BUNDLE})
        self.save_element(prov_element=bundle_record)

        return self._save_bundle_internal(prov_bundle)

    def _save_bundle_internal(self, prov_bundle):
        """
        Private method to create a bundle in the database

        :param prov_bundle: ProvBundle
        :type prov_bundle: prov.model.ProvBundle
        :return bundle_id: The bundle from the database adapter
        :rtype: str
        """
        if not isinstance(prov_bundle, ProvBundle):
            raise InvalidArgumentTypeException()

        bundle_id = str(uuid4())
        # create nodes
        for record in prov_bundle.get_records(ProvElement):
            self.save_element(prov_element=record, bundle_id=bundle_id)

        # create relations
        for relation in prov_bundle.get_records(ProvRelation):

            self.save_relation(relation, bundle_id)

        return bundle_id

    def save_relation(self, prov_relation, bundle_id=None):
        """
        Saves a relation between 2 nodes that are already in the database.

        .. code:: python

            doc = ProvDocument()

            activity    = doc.activity("ex:yourActivity")
            entity      = doc.entity("ex:yourEntity")
            wasGeneratedBy = entity.wasGeneratedBy("ex:yourAgent")

            # Save the elements
            rel_id = prov_db.save_relation(wasGeneratedBy)


        :param prov_relation: The ProvRelation instance
        :type prov_relation: ProvRelation
        :param bundle_id
        :type bundle_id: str
        :return: Relation id
        :rtype: str
        """

        if not isinstance(prov_relation, ProvRelation):
            raise InvalidArgumentTypeException(
                "prov_relation was {}, expected: {}".format(type(prov_relation), type(ProvRelation)))


        # get from and to node
        from_tuple, to_tuple = prov_relation.formal_attributes[:2]
        from_qualified_name = from_tuple[1]
        to_qualified_name = to_tuple[1]

        # if target or origin record is unknown, save node "Unknown"
        if from_qualified_name is None:
            uid = uuid4()
            from_qualified_name = prov_relation.bundle.valid_qualified_name("prov:Unknown-{}".format(uid))
            del uid

        if to_qualified_name is None:
            uid = uuid4()
            to_qualified_name = prov_relation.bundle.valid_qualified_name("prov:Unknown-{}".format(uid))
            del uid

        # Ensure that the from and to node exists
        relation_cls = PROV_REC_CLS[prov_relation.get_type()]
        from_type,to_type = relation_cls.FORMAL_ATTRIBUTES[:2]

        # get the class types
        from_type_cls = PROV_ATTR_BASE_CLS[from_type]
        to_type_cls = PROV_ATTR_BASE_CLS[to_type]

        if from_type_cls is None or to_type_cls is None:
            raise InvalidArgumentTypeException(
                "Could not determinate typ for relation from: {}, to: {}, prov_relation was {}, ".format(from_type, to_type, type(prov_relation)))
        #save from and to node
        self.save_element(prov_element=from_type_cls(prov_relation.bundle, identifier=from_qualified_name), bundle_id=bundle_id)

        to_bundle = prov_relation.bundle

        # If it is a link between bundle the to node not belongs to the current bundle, the to node belongs only to the bundle defined as FORMAL_ATTR[3]
        if prov_relation.get_type() is PROV_MENTION:

            #Try to get the destination bundle
            to_bundle_identifier = list(prov_relation.get_attribute(PROV_ATTR_BUNDLE)).pop()

            if not isinstance(to_bundle_identifier,QualifiedName):
                raise InvalidProvRecordException("Should be a qualified name {}, mention: {}".format(to_bundle_identifier, prov_relation))
            # Create the bundle, it will be automatically created during the save_element method
            to_bundle = ProvBundle(identifier=to_bundle_identifier)


        self.save_element(prov_element=to_type_cls(to_bundle, identifier=to_qualified_name), bundle_id=bundle_id)


        # split metadata and attributes
        (metadata, attributes) = self._get_metadata_and_attributes_for_record(prov_relation)

        # Include namespace uri into the identifier to support e.g. different default namespaces
        global_from_qualified_name = from_qualified_name.namespace.uri + from_qualified_name.localpart
        global_to_qualified_name = to_qualified_name.namespace.uri + to_qualified_name.localpart

        return self._adapter.save_relation(global_from_qualified_name, global_to_qualified_name,
                                           attributes, metadata)

    def _create_bundle_association(self, prov_elements, prov_bundle_identifier):
        """
        This method saves a relation between the bundle entity and all nodes in the bundle

        :param prov_bundle_identifier: The bundle identifier
        :type prov_bundle_identifier: QualifiedName
        :param prov_elements: List of prov elements
        :type prov_elements: list
        """

        # Ensure that the bundle entity exist
        doc = ProvDocument()
        to_bundle = ProvBundle(document=doc,identifier=prov_bundle_identifier)
        self.save_bundle(to_bundle) # Save the empty bundle to create the bundle entity if necessary

        belong_relation = ProvAssociation(bundle=to_bundle, identifier=None,
                                          attributes={PROV_TYPE: "prov:bundleAssociation"})
        (belong_metadata, belong_attributes) = self._get_metadata_and_attributes_for_record(belong_relation)
        to_qualified_name = prov_bundle_identifier

        for record in prov_elements:
            (metadata, attributes) = self._get_metadata_and_attributes_for_record(record)
            from_qualified_name = metadata[METADATA_KEY_IDENTIFIER]
            global_prov_to_identifier= to_qualified_name.namespace.uri + to_qualified_name.localpart
            self._adapter.save_relation(from_qualified_name, global_prov_to_identifier,
                                        belong_attributes, belong_metadata)


    def _save_bundle_links(self, prov_bundle):
        """
        This function saves the links between nodes in bundles, see https://www.w3.org/TR/prov-links/

        :param prov_bundle: For this bundle we will create the links
        :type prov_bundle: ProvBundle
        """

        for mention in prov_bundle.get_records(ProvRelation):
            if mention.get_type() is not PROV_MENTION:
                continue

            self.save_relation(mention)

    @staticmethod
    def _get_metadata_and_attributes_for_record(prov_record, bundle_id=None):
        """
        This function generate some meta data for the record for example:

            * Namespaces: The prov_record use several namespaces and the metadata contain this namespaces
            * Type_Map: The type map is important to get exactly the same document back, you have to save this information (like what attribute is a datetime)

        :param prov_record: The ProvRecord (ProvRelation or ProvElement)
        :type prov_record: ProvRecord
        :param bundle_id: The id of the document
        :type bundle_id: str
        :return: namedtuple(metadata, attributes)
        :rtype: namedtuple
        """
        if not isinstance(prov_record, ProvRecord):
            raise InvalidArgumentTypeException()

        used_namespaces = dict()
        bundle = prov_record.bundle

        prov_type = prov_record.get_type()
        prov_identifier = prov_record.identifier

        if prov_type is None and isinstance(prov_record, ProvRecord):
            prov_type = bundle.valid_qualified_name("prov:Unknown")

        # if relation without identifier -> use prov_type as identifier
        if prov_identifier is None and prov_record.identifier is None:
            prov_identifier = prov_type

        # Be sure that the prov_identifier is a qualified name instance

        if not isinstance(prov_identifier, QualifiedName):
            qualified_name = bundle.valid_qualified_name(prov_identifier)
            if qualified_name is None:
                raise InvalidProvRecordException(
                    "The prov record {} is invalid because the prov_identifier {} can't be qualified".format(
                        prov_record, prov_identifier))
            else:
                prov_identifier = qualified_name

        # extract namespaces from record

        # add namespace from prov_type
        namespace = prov_type.namespace
        used_namespaces.update(serialize_namespace(namespace))

        # add namespace from prov identifier
        namespace = prov_identifier.namespace
        used_namespaces.update(serialize_namespace(namespace))

        attributes = dict(prov_record.attributes.copy())
        for key, value in attributes.items():

            # ensure key is QualifiedName
            if isinstance(key, QualifiedName):
                namespace = key.namespace
                used_namespaces.update(serialize_namespace(namespace))
            else:
                raise InvalidProvRecordException("Not support key type {}".format(type(key)))

            # try to add
            if isinstance(value, QualifiedName):
                namespace = value.namespace
                used_namespaces.update(serialize_namespace(namespace))
            else:
                qualified_name = bundle.valid_qualified_name(value)
                if qualified_name is not None:
                    # Don't update the attribute, so we only save the namespace instead of the attribute as a qualified name.
                    # For some reason the prov-library allow a string with a schema: <namespace_prefix>:<identifier>
                    # This line cause an error during the test: "test_primer_example_alternate"
                    # attributes[key] = qualified_name # update attribute

                    namespace = qualified_name.namespace
                    used_namespaces.update(serialize_namespace(namespace))

        # create type dict
        types_dict = dict()
        for key, value in attributes.items():
            if key not in PROV_ATTRIBUTES:
                return_type = encode_json_representation(value)
                if return_type is not None:
                    types_dict.update({str(key): return_type})

        # Include namespace uri into the identifier to support e.g. different default namespaces
        global_prov_identifier = prov_identifier.namespace.uri + prov_identifier.localpart

        metadata = {
            METADATA_KEY_PROV_TYPE: prov_type,
            METADATA_KEY_IDENTIFIER: global_prov_identifier,
            METADATA_KEY_IDENTIFIER_ORIGINAL: prov_identifier,
            METADATA_KEY_NAMESPACES: used_namespaces,
            METADATA_KEY_TYPE_MAP: types_dict
        }

        # Add document id to metadata, to restore the
        if bundle_id:
            metadata.update({bundle_id: True})
        meta_and_attributes = namedtuple("MetaAndAttributes", "metadata, attributes")

        return meta_and_attributes(metadata, attributes)



#!/usr/bin/env python
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

tests_require = [
    'coverage',
    'coveralls'
]

docs_require = [
    'Sphinx>=1.3.5',
    'recommonmark>=0.4.0',
    'sphinx-rtd-theme>=0.1.9',
    'sphinxcontrib-napoleon>=0.4.4',
    'sphinxcontrib-httpdomain>=1.5.0',
    'mock',
]

setup(
    name='prov-db-connector',
    version='0.5.1',
    description='PROV Database Connector',
    keywords=[
        'provenance', 'graph', 'model', 'PROV', 'PROV-DM', 'PROV-JSON', 'JSON',
        'PROV-XML', 'PROV-N'
    ],
    author='DLR, Stefan Bieliauskas, Martin Stoffers',
    author_email='opensource@dlr.de, sb@conts.de, martin.stoffers@studserv.uni-leipzig.de',
    url='https://github.com/DLR-SC/prov-db-connector',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    license="Apache License 2.0",

    packages=find_packages(),
    package_dir={
        'provdbconnector': 'provdbconnector'
    },
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "prov==2.0.0",
        "neo4j==4.4.1"
    ],
    extras_require={
        'test': tests_require,
        'dev': tests_require + docs_require,
        'docs': docs_require,
    },

    test_suite='provdbconnector.tests',
)



from provdbconnector.exceptions.database import InvalidOptionsException
from provdbconnector.db_adapters.in_memory import SimpleInMemoryAdapter
from provdbconnector.prov_db import ProvDb
from provdbconnector.tests import AdapterTestTemplate
from provdbconnector.tests import ProvDbTestTemplate


class SimpleInMemoryAdapterTest(AdapterTestTemplate):
    """
    This class implements the AdapterTestTemplate and only override some functions.

    """
    def setUp(self):
        """
        Connect to your database

        """
        self.instance = SimpleInMemoryAdapter()
        self.instance.connect(None)

    def test_connect_invalid_options(self):
        """
        Test your connect function with invalid data

        """
        auth_info = {"invalid": "Invalid"}
        with self.assertRaises(InvalidOptionsException):
            self.instance.connect(auth_info)

    def clear_database(self):
        """
        Clear the database

        """
        self.instance.all_nodes = dict()
        self.instance.all_relations= dict()

    def tearDown(self):
        """
        Delete your instance

        """
        del self.instance


class SimpleInMemoryAdapterProvDbTests(ProvDbTestTemplate):
    """
    This is the high level test for the SimpleInMemoryAdapter

    """
    def setUp(self):
        """
        Setup a ProvDb instance
        """
        self.provapi = ProvDb(api_id=1, adapter=SimpleInMemoryAdapter, auth_info=None)

    def clear_database(self):
        """
        Clear function get called before each test starts

        """
        self.provapi._adapter.all_nodes = dict()
        self.provapi._adapter.all_relations = dict()

    def tearDown(self):
        """
        Delete prov api instance
        """
        del self.provapi



import logging
from collections import namedtuple

log = logging.getLogger(__name__).addHandler(logging.NullHandler())

METADATA_PARENT_ID = "parent_id"
METADATA_KEY_PROV_TYPE = "prov_type"
METADATA_KEY_IDENTIFIER = "identifier"
METADATA_KEY_IDENTIFIER_ORIGINAL = "identifier_original"
METADATA_KEY_NAMESPACES = "namespaces"
METADATA_KEY_TYPE_MAP = "type_map"

# Return types for adapter classes
DbDocument = namedtuple("DbDocument", "document, bundles")
DbBundle = namedtuple("DbBundle", "records, bundle_record")

DbRecord = namedtuple("DbRecord", "attributes, metadata")
DbRelation = namedtuple("DbRelation", "attributes, metadata")


class BaseAdapter():
    """
    Interface class for a prov database adapter
    """

    def __init__(self, *args, **kwargs):
        pass

    def connect(self, authentication_info):
        """
        Establish the database connection / login into the database

        :param authentication_info: a custom dict with credentials
        :type authentication_info: dict
        :return: Indicate whether the connection was successful
        :rtype: boolean
        :raise InvalidOptionsException:
        """
        raise NotImplementedError("Abstract method")

    def save_element(self, attributes, metadata):
        """
        Saves a entity, activity or entity into the database

        :param attributes: Attributes as dict for the record. Be careful you have to encode the dict
        :type attributes: dict
        :param metadata: Metadata as dict for the record. Be careful you have to encode the dict but you can be sure that all meta keys are always there
        :type metadata: dict
        :return: Record id
        :rtype: str
        """
        raise NotImplementedError("Abstract method")

    def save_relation(self, from_node, to_node, attributes, metadata):
        """
        Create a relation between 2 nodes

        :param from_node: The identifier
        :type from_node: str
        :param to_node: The identifier for the destination node
        :type: to_node: str
        :param attributes:  Attributes as dict for the record. Be careful you have to encode the dict
        :type attributes: dict
        :param metadata: Metadata as dict for the record. Be careful you have to encode the dict but you can be sure that all meta keys are always there
        :type metadata: dict
        :return: Record id
        :rtype: str
        """
        raise NotImplementedError("Abstract method")

    def get_records_by_filter(self, attributes_dict=None, metadata_dict=None):
        """
        Returns all records (nodes and relations) based on a filter dict.
        The filter dict's are and AND combination but only the start node must fulfill the conditions.
        The result should contain all associated relations and nodes together

        :param attributes_dict:
        :type attributes_dict: dict
        :param metadata_dict:
        :type metadata_dict: dict
        :return: list of relations and nodes
        :rtype: list
        """
        raise NotImplementedError("Abstract method")

    def get_records_tail(self, attributes_dict=None, metadata_dict=None, depth=None):
        """
        Returns all connected nodes and relations based on a filter.
        The filter is an AND combination and this describes the filter only for the origin nodes.

        :param attributes_dict:
        :type attributes_dict: dict
        :param metadata_dict:
        :type metadata_dict: dict
        :param depth:
        :type depth: int
        :return: a list of relations and nodes
        :rtype: list
        """
        raise NotImplementedError("Abstract method")

    def get_bundle_records(self, bundle_identifier):
        """
        Returns the relations and nodes for a specific bundle identifier.
        Please use the bundle association to get all bundle nodes.
        Only the relations belongs to the bundle where the start AND end node belong also to the bundle.
        Except the prov:Mention see: W3C bundle links

        :param bundle_identifier: The bundle identifier
        :type bundle_identifier: str
        :return: list of nodes and bundles
        :rtype: list
        """
        raise NotImplementedError("Abstract method")

    def get_record(self, record_id):
        """
        Return a single record

        :param record_id: The id
        :type record_id: str
        :return: DbRecord
        :rtype: DbRecord
        """
        raise NotImplementedError("Abstract method")

    def get_relation(self, relation_id):
        """
        Returns a single relation

        :param relation_id: The id
        :type relation_id: str
        :return: DbRelation
        :rtype: DbRelation
        """
        raise NotImplementedError("Abstract method")

    def delete_records_by_filter(self, attributes_dict, metadata_dict):
        """
        Delete records by filter

        :param attributes_dict:
        :type attributes_dict: dict
        :param metadata_dict:
        :type metadata_dict: dict
        :return: Indicates whether the deletion was successful
        :rtype: boolean
        :raise NotFoundException:
        """
        raise NotImplementedError("Abstract method")

    def delete_record(self, record_id):
        """
        Delete a single record

        :param record_id:
        :type record_id: str
        :return: Indicates whether the deletion was successful
        :rtype: boolean
        :raise NotFoundException:
        """
        raise NotImplementedError("Abstract method")

    def delete_relation(self, relation_id):
        """
        Delete a single relation

        :param relation_id:
        :type relation_id: str
        :return: Indicates whether the deletion was successful
        :rtype: boolean
        :raise NotFoundException:
        """
        raise NotImplementedError("Abstract method")



import json
import logging
import sys
from collections import namedtuple
from datetime import datetime
from io import StringIO

import six
from prov.constants import PROV_QUALIFIEDNAME, PROV_ATTRIBUTES_ID_MAP, PROV_ATTRIBUTES, PROV_MEMBERSHIP, \
    PROV_ATTR_ENTITY, PROV_ATTRIBUTE_QNAMES, PROV_ATTR_COLLECTION, XSD_ANYURI, PROV_ATTR_ACTIVITY, PROV_ATTR_AGENT, \
    PROV_ATTR_TRIGGER,PROV_ATTR_INFORMED,PROV_ATTR_INFORMANT,PROV_ATTR_STARTER,PROV_ATTR_ENDER,PROV_ATTR_AGENT \
    ,PROV_ATTR_PLAN,PROV_ATTR_DELEGATE,PROV_ATTR_RESPONSIBLE,PROV_ATTR_GENERATED_ENTITY,PROV_ATTR_USED_ENTITY, \
    PROV_ATTR_GENERATION,PROV_ATTR_USAGE,PROV_ATTR_SPECIFIC_ENTITY,PROV_ATTR_GENERAL_ENTITY,PROV_ATTR_ALTERNATE1, \
    PROV_ATTR_ALTERNATE2,PROV_ATTR_BUNDLE,PROV_ATTR_INFLUENCEE,PROV_ATTR_INFLUENCER

from prov.model import Literal, Identifier, QualifiedName, Namespace, parse_xsd_datetime, PROV_REC_CLS, ProvAgent, \
    ProvEntity, ProvActivity, ProvElement
from provdbconnector.db_adapters.baseadapter import METADATA_KEY_NAMESPACES, METADATA_KEY_PROV_TYPE, \
    METADATA_KEY_TYPE_MAP
from provdbconnector.exceptions.database import MergeException
from provdbconnector.exceptions.provapi import InvalidArgumentTypeException
from provdbconnector.exceptions.utils import SerializerException

log = logging.getLogger(__name__)

logger = logging.getLogger(__name__)

FormalAndOtherAttributes = namedtuple("formal_and_other_attributes", "formal, other")

# Reverse map for prov.model.XSD_DATATYPE_PARSERS
LITERAL_XSDTYPE_MAP = {
    float: 'xsd:double',
    int: 'xsd:int'
    # boolean, string values are supported natively by PROV-JSON
    # datetime values are converted separately
}

# Add long on Python 2
if six.integer_types[-1] not in LITERAL_XSDTYPE_MAP:
    LITERAL_XSDTYPE_MAP[six.integer_types[-1]] = 'xsd:long'


PROV_ATTR_BASE_CLS = {
    PROV_ATTR_ENTITY: ProvEntity,
    PROV_ATTR_ACTIVITY: ProvActivity,
    PROV_ATTR_TRIGGER: ProvEntity,
    PROV_ATTR_INFORMED: ProvActivity,
    PROV_ATTR_INFORMANT: ProvActivity,
    PROV_ATTR_STARTER: ProvActivity,
    PROV_ATTR_ENDER: ProvActivity,
    PROV_ATTR_AGENT: ProvAgent,
    PROV_ATTR_PLAN: ProvEntity,
    PROV_ATTR_DELEGATE: ProvAgent,
    PROV_ATTR_RESPONSIBLE: ProvAgent,
    PROV_ATTR_GENERATED_ENTITY: ProvEntity,
    PROV_ATTR_USED_ENTITY: ProvEntity,
    PROV_ATTR_GENERATION: ProvElement,
    PROV_ATTR_USAGE: ProvElement,
    PROV_ATTR_SPECIFIC_ENTITY: ProvEntity,
    PROV_ATTR_GENERAL_ENTITY: ProvEntity,
    PROV_ATTR_ALTERNATE1: ProvEntity,
    PROV_ATTR_ALTERNATE2: ProvEntity,
    PROV_ATTR_BUNDLE:ProvElement,
    PROV_ATTR_INFLUENCEE: ProvElement,
    PROV_ATTR_INFLUENCER: ProvElement,
    PROV_ATTR_COLLECTION: ProvEntity
}

def encode_dict_values_to_primitive(dict_values):
    """
    This function transforms a dict with all kind of types into a dict with only

    - str
    - dict
    - book
    - str

    values

    :param dict_values:
    :return:
    """
    new_dict_values = dict()
    for key, value in dict_values.items():
        key_simple = str(key)
        new_dict_values.update({key_simple: encode_string_value_to_primitive(value)})

    return new_dict_values


def encode_string_value_to_primitive(value):
    """
    Convert a value into one of the following types:

    - dict
    - str
    - float
    - int
    - list


    :param value:
    :return:
    """
    if isinstance(value, Literal):
        return value.value
    elif type(value) is int:
        return value
    elif type(value) is float:
        return value
    elif type(value) is bool:
        return value
    elif type(value) is list:
        return value
    elif type(value) is dict:
        io = StringIO()
        json.dump(value, io)
        return io.getvalue()
    return str(value)


def literal_json_representation(literal):
    """
    Some internationalization stuff

    :param literal:
    :return:
    """
    # TODO: QName export
    value, datatype, langtag = literal.value, literal.datatype, literal.langtag
    if langtag:
        return {'lang': langtag}
    else:
        return {'type': six.text_type(datatype)}


def encode_json_representation(value):
    """
    Get the type of a value

    :param value:
    :return:
    """
    if isinstance(value, Literal):
        return literal_json_representation(value)
    elif isinstance(value, datetime):
        return {'type': 'xsd:dateTime'}
    elif isinstance(value, QualifiedName):
        # TODO Manage prefix in the whole structure consistently
        # TODO QName export
        return {'type': str(PROV_QUALIFIEDNAME)}
    elif isinstance(value, Identifier):
        return {'type': 'xsd:anyURI'}
    elif type(value) in LITERAL_XSDTYPE_MAP:
        return {'type': LITERAL_XSDTYPE_MAP[type(value)]}
    else:
        return None


# DECODE

def add_namespaces_to_bundle(prov_bundle, metadata):
    """
    Add all namespaces in the metadata_dict to the provided bundle

    :param prov_bundle:
    :param metadata:
    :return: None
    """
    namespaces = dict()
    try:
        namespace_str = metadata[METADATA_KEY_NAMESPACES]
    except ValueError:
        SerializerException("No valid namespace provided, should be a string of a dict: {}".format(metadata))
        return

    if type(namespace_str) is str:
        io = StringIO(namespace_str)
        namespaces = json.load(io)
    elif type(namespace_str) is dict:
        namespaces = namespace_str
    elif type(namespace_str) is list:
        for entry in namespace_str:
            if type(entry) is str:
                io = StringIO(entry)
                namespaces.update(json.load(io))
            else:
                raise SerializerException(
                    "Namespaces metadata should returned as json string dict or list of json strings not as {}".format(
                        type(namespace_str)))

    else:
        raise SerializerException(
            "Namespaces metadata should returned as json string dict or list of json strings not as  {}".format(
                type(namespace_str)))

    for prefix, uri in namespaces.items():
        if prefix is not None and uri is not None:
            if prefix != 'default':
                prov_bundle.add_namespace(Namespace(prefix, uri))
            else:
                prov_bundle.set_default_namespace(uri)
        else:
            SerializerException("No valid namespace provided for the metadata: {}".format(metadata))


def create_prov_record(bundle, prov_type, prov_id, properties, type_map):
    """

    :param bundle:
    :param prov_type: valid prov type like prov:Entry as string
    :param prov_id: valid id as string like <namespace>:<name>
    :param properties: dict{attr_name:attr_value} dict with all properties (prov and additional)
    :param type_map: dict{attr_name:type_str} Contains the type information for each property (only if type is necessary)
    :return: ProvRecord
    """
    # Parse attributes
    if isinstance(properties, dict):
        properties_list = properties.items()
    elif isinstance(properties, list):
        properties_list = properties
    else:
        raise SerializerException(
            "Please provide properties as list[(key,value)] or dict your provided: {}".format(
                properties.__class__.__name__))

    attributes = dict()
    other_attributes = []
    # this is for the multiple-entity membership hack to come
    membership_extra_members = None
    for attr_name, values in properties_list:

        attr = (
            PROV_ATTRIBUTES_ID_MAP[attr_name]
            if attr_name in PROV_ATTRIBUTES_ID_MAP
            else bundle.valid_qualified_name(attr_name)
        )
        if attr in PROV_ATTRIBUTES:
            if isinstance(values, list):
                # only one value is allowed
                if len(values) > 1:
                    # unless it is the membership hack
                    if prov_type == PROV_MEMBERSHIP and \
                                    attr == PROV_ATTR_ENTITY:
                        # This is a membership relation with
                        # multiple entities
                        # HACK: create multiple membership
                        # relations, one x each entity

                        # Store all the extra entities
                        membership_extra_members = values[1:]
                        # Create the first membership relation as
                        # normal for the first entity
                        value = values[0]
                    else:
                        error_msg = (
                            'The prov package does not support PROV'
                            ' attributes having multiple values.'
                        )
                        logger.error(error_msg)
                        raise SerializerException(error_msg)
                else:
                    value = values[0]
            else:
                value = values

            if attr in PROV_ATTRIBUTE_QNAMES:
                value = (bundle.valid_qualified_name(value))
            elif isinstance(value, datetime):
                value = value
            else:
                parse_xsd_datetime(value)

            attributes[attr] = value
        else:
            value_type = None
            if type_map:
                value_type = type_map.get(attr_name)

            if isinstance(values, list):
                other_attributes.extend(
                    (
                        attr,
                        decode_json_representation(value, value_type, bundle)
                    )
                    for value in values
                )
            else:
                # single value
                other_attributes.append(
                    (
                        attr,
                        decode_json_representation(values, value_type, bundle)
                    )
                )
    record = bundle.new_record(
        prov_type, prov_id, attributes, other_attributes
    )
    # HACK: creating extra (unidentified) membership relations
    if membership_extra_members:
        collection = attributes[PROV_ATTR_COLLECTION]
        for member in membership_extra_members:
            bundle.membership(
                collection, bundle.valid_qualified_name(member)
            )
    return record


def decode_json_representation(value, type, bundle):
    """
    Return the value based on the type see also encode_json_representation

    :param value:
    :param type:
    :param bundle:
    :return:
    """
    if isinstance(type, dict):
        # complex type
        datatype = type['type'] if 'type' in type else None
        datatype = bundle.valid_qualified_name(datatype)
        langtag = type['lang'] if 'lang' in type else None
        if datatype == XSD_ANYURI:
            return Identifier(value)
        elif datatype == PROV_QUALIFIEDNAME:
            return bundle.valid_qualified_name(value)
        else:
            # The literal of standard Python types is not converted here
            # It will be automatically converted when added to a record by
            # _auto_literal_conversion()
            return Literal(value, datatype, langtag)
    else:
        # simple type, just return it
        return value


def split_into_formal_and_other_attributes(attributes, metadata):
    """
    This function split the attributes and metadata into formal attributes and other attributes.
    Helpful for merge operations and searching for duplicate relations

    :param attributes:
    :param metadata:
    :return: namedtuple(formal_attributes, other_attributes)
    :rtype: FormalAndOtherAttributes
    """
    prov_type = metadata[METADATA_KEY_PROV_TYPE]

    if str(prov_type) == "prov:Unknown":
        formal_qualified_names = set()
    else:
        class_type = PROV_REC_CLS[prov_type]
        formal_qualified_names = class_type.FORMAL_ATTRIBUTES

    formal_attributes = {key: attributes[key] for key in attributes.keys() if key in formal_qualified_names}
    other_attributes = {key: attributes[key] for key in attributes.keys() if key not in formal_qualified_names}

    return FormalAndOtherAttributes(formal_attributes, other_attributes)

def merge_record(attributes, metadata, other_attributes, other_metadata):
    """
    Merge 2 records into one

    :param attributes: The original attributes
    :param metadata: The original metadata
    :param other_attributes: The attributes to merge
    :param other_metadata:  The metadata to merge
    :return: tuple(attributes, metadata)
    :rtype: Tuple(attributes,metadata)
    """
    attributes_merged = attributes.copy()
    attributes_merged.update(other_attributes)

    metadata_prov_typ = metadata[METADATA_KEY_PROV_TYPE]
    other_metadata_prov_typ = other_metadata[METADATA_KEY_PROV_TYPE]

    # Determinate non unknown prov type for merge
    merged_prov_typ = other_metadata_prov_typ
    is_one_prov_type_unknown = False
    # Support unknown typ during the merge process, needed for polymorph nodes
    if other_metadata_prov_typ.localpart == "Unknown":
        merged_prov_typ = metadata_prov_typ
        is_one_prov_type_unknown = True

    if metadata_prov_typ.localpart == "Unknown":
        merged_prov_typ = other_metadata_prov_typ
        is_one_prov_type_unknown = True

    if merged_prov_typ.localpart == "Unknown":
        raise MergeException("Prov type can't be unknown metadata: {}, other metadata: {}".format(metadata_prov_typ, other_metadata_prov_typ))

    if metadata_prov_typ != other_metadata_prov_typ and not is_one_prov_type_unknown:
        raise MergeException(
            "Prov type should be the same but is: {}:{}".format(metadata_prov_typ, other_metadata_prov_typ))

    for (key, value) in attributes.items():
        if attributes_merged[key] != value:
            raise MergeException(
                "Invalid data, it is not allowed to override existing attributes key:{}, value:{} with value:{}".format(
                    key, value, attributes_merged[key]))

    merged_metadata_namespaces = metadata[METADATA_KEY_NAMESPACES].copy()
    merged_metadata_namespaces.update(other_metadata[METADATA_KEY_NAMESPACES])

    merged_metadata_type_map = metadata[METADATA_KEY_TYPE_MAP].copy()
    merged_metadata_type_map.update(other_metadata[METADATA_KEY_TYPE_MAP])

    merged_metadata = metadata.copy()
    merged_metadata.update(other_metadata)
    merged_metadata.update({METADATA_KEY_PROV_TYPE: merged_prov_typ})
    merged_metadata.update({METADATA_KEY_NAMESPACES: merged_metadata_namespaces})
    merged_metadata.update({METADATA_KEY_TYPE_MAP: merged_metadata_type_map})

    return attributes_merged, merged_metadata

def serialize_namespace(namespace: Namespace):
    prefix = namespace.prefix

    if prefix == "" or prefix is None:
        prefix = "default"
    return {str(prefix): str(namespace.uri)}
